// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "dawnlibs.hpp"
#include "assert/assert.hpp"
#include "util/memory.hpp"

namespace Dawn {
  class AssetLoader {
    private:
      std::string fileName;
      FILE *handle;

    public:
      /**
       * Create a new asset loader. Asset Loaders can be used to load data from
       * a file in a myriad of ways.
       * 
       * @param fileName File name of the asset that is to be loaded.
       */
      AssetLoader(std::string fileName);

      /**
       * Platform-centric method to open a file buffer to an asset.
       * 
       * @return 0 if success, otherwise for failure.
       */
      void open();

      /**
       * Closes the previously ppened asset.
       * @return 0 if successful, otherwise false.
       */
      int32_t close();

      /**
       * Read bytes from buffer.
       * @param buffer Pointer to a ubyte array to buffer data into.
       * @param size Length of the data buffer (How many bytes to read).
       * @return The count of bytes read.
       */
      size_t read(uint8_t *buffer, size_t size);

      /**
       * Skip to the end of the buffer, useful to find the length of the buffer.
       * @return 0 if successful, otherwise false.
       */
      int32_t end();

      /**
       * Method to skip n bytes in the buffer
       * @param n Count of bytes to skip.
       * @return 0 if successful, otherwise unsuccessful.
       */
      size_t skip(size_t n);

      /**
       * Rewinds to the start of the asset buffer.
       * @return 0 if successful, otherwise unsuccessful.
       */      
      int32_t rewind();
      
      /**
       * Retreive the current byte position within the asset that the head is
       * at.
       * @return Position (in bytes) that the current seek is at.
       */
      size_t getPosition();

      /**
       * Sets the position of the file read head.
       * 
       * @param position Position to set to.
       * @return How many bytes were skipped / rewound.
       */
      size_t setPosition(size_t position);

      /**
       * Loads the entire file into a raw buffer.
       * @param buffer Pointer to where a pointer to the buffer will be stored.
       * @return Size of the buffer that was read (in bytes).
       */
      size_t loadRaw(uint8_t **buffer);

      /**
       * Run a callback for each byte within the asset. The callback will
       * receive each byte individually.
       * 
       * @tparam T Type of instance to run callback against.
       * @param instance Instance of the object to run the callback against.
       * @param callback Callback method on the class to run the callback for.
       * @return The count of bytes read.
       */
      template<class T>
      size_t loadBufferedCallback(T *instance, bool (T::*callback)(uint8_t n)) {
        uint8_t buffer[1024];
        size_t read, length;
        int32_t i;
        bool result;

        assertNotNull(instance);
        assertNotNull(callback);

        // Open the buffer.
        this->open();

        // Reset length size
        length = 0;

        // Buffer from input
        while((read = this->read(buffer, 1024)) != 0) {
          for(i = 0; i < read; i++) {
            result = ((*instance).*(callback))(buffer[i]);
            if(!result) {
              length += i;
              break;
            }
          }
          if(!result) break;
          length += read;
        }

        // Close the buffer
        this->close();

        return length;
      }

      /**
       * Cleanup the asset loader.
       */
      virtual ~AssetLoader();
  };
}